//******************************************************************************
///
/// @file parser/parser_strings.cpp
///
/// This module implements parsing and conversion of string expressions.
///
/// @copyright
/// @parblock
///
/// Persistence of Vision Ray Tracer ('POV-Ray') version 3.8.
/// Copyright 1991-2023 Persistence of Vision Raytracer Pty. Ltd.
///
/// POV-Ray is free software: you can redistribute it and/or modify
/// it under the terms of the GNU Affero General Public License as
/// published by the Free Software Foundation, either version 3 of the
/// License, or (at your option) any later version.
///
/// POV-Ray is distributed in the hope that it will be useful,
/// but WITHOUT ANY WARRANTY; without even the implied warranty of
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
/// GNU Affero General Public License for more details.
///
/// You should have received a copy of the GNU Affero General Public License
/// along with this program.  If not, see <http://www.gnu.org/licenses/>.
///
/// ----------------------------------------------------------------------------
///
/// POV-Ray is based on the popular DKB raytracer version 2.12.
/// DKBTrace was originally written by David K. Buck.
/// DKBTrace Ver 2.0-2.12 were written by David K. Buck & Aaron A. Collins.
///
/// @endparblock
///
//------------------------------------------------------------------------------
// SPDX-License-Identifier: AGPL-3.0-or-later
//******************************************************************************

// Unit header file must be the first file included within POV-Ray *.cpp files (pulls in config)
#include "parser/parser.h"

// C++ variants of C standard header files
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <cstring>

// C++ standard header files
//  (none at the moment)

// Boost header files
#include <boost/date_time/posix_time/posix_time.hpp>

// POV-Ray header files (base module)
#include "base/pov_mem.h"
#include "base/stringutilities.h"

// POV-Ray header files (core module)
#include "core/scene/scenedata.h"

// POV-Ray header files (parser module)
//  (none at the moment)

// this must be the last file included
#include "base/povdebug.h"

namespace pov_parser
{

using namespace pov;

// Array used with keyword fork_str(n), FORK_STR_TOKEN, Parse_ForkStr()
// NB. Keep this aligned with fork_str(n), FORK_VAL_TOKEN, Fork_Values[]
// NB. PACKAGE_VERSION macro is already defined as a string.
static char *Fork_Strings[FORK_STR_FORK_VAL_MAX_COUNT] = {
    "povr",
    PACKAGE_VERSION,
    "Branched off POV-Ray master at commit 74b3ebe0. March 19, 2019.",
    "Major",
    "Minor",
    "Revision",
    "Patch"
};

// These static arrays created with the help of code scanning tools. See /home/pokorny/bin/word_is
// This set must match that in parser_expressions.cpp. The iFnNames also one in parser.cpp. TODO. Fix. (optimized)
static char iKeyWords[][26] { "aa_level", "aa_threshold", "abs", "absorption", "accuracy", "acos", "acosh", "adaptive",
    "adc_bailout", "agate", "albedo", "all", "all_intersections", "altitude", "always_sample", "ambient", "amplify",
    "angle", "anisotropy", "aoi", "aperture", "append", "arc_angle", "area_illumination", "area_light", "array", "asc",
    "ascii", "asin", "asinh", "assumed_gamma", "atan", "atan2", "atanh", "autostop", "average", "b_spline",
    "background", "bevy", "bezier_spline", "bicubic_patch", "bitwise_and", "bitwise_or", "bitwise_xor", "black_hole",
    "blend_gamma", "blend_mode", "blob", "blue", "blur_samples", "bmp", "bokeh", "bounded_by", "box", "boxed", "bozo",
    "brick", "brick_size", "brightness", "brilliance", "bt2020", "bt709", "bump_map", "bump_size", "bumps", "camera",
    "caustics", "ceil", "cells", "charset", "checker", "chr", "circular", "clipped_by", "clock", "clock_on", "cmap",
    "collect", "color", "color_map", "colour", "colour_map", "component", "concat", "cone", "confidence", "conic_sweep",
    "conserve_energy", "contained_by", "control0", "control1", "coords", "cos", "cosh", "count", "crackle", "crand",
    "cube", "cubic", "cubic_spline", "cubic_wave", "cutaway_textures", "cylinder", "cylindrical", "datetime", "default",
    "defined", "degrees", "density", "density_file", "density_map", "dents", "deprecated", "df3", "dictionary",
    "difference", "diffuse", "dimension_size", "dimensions", "direction", "disc", "dispersion", "dispersion_samples",
    "dist_exp", "distance", "div", "double_illuminate", "e", "eccentricity", "emission", "error_bound", "exp",
    "expand_thresholds", "exponent", "exr", "exterior", "extinction", "face_indices", "facets", "fade_color",
    "fade_colour", "fade_distance", "fade_power", "falloff", "falloff_angle", "false", "fclose", "file_exists",
    "filter", "finish", "fisheye", "flatness", "flip", "floor", "focal_point", "fog", "fopen", "fork_str", "fork_val",
    "form", "frequency", "fresnel", "function", "function_interval", "gamma", "gather", "getenv", "gif", "global",
    "global_lights", "global_settings", "gradient", "granite", "gray", "gray_threshold", "green", "grey",
    "grey_threshold", "grooved", "hard_object", "hdr", "height_field", "hexagon", "hi", "hierarchy", "hollow",
    "hypercomplex", "id_type", "iff", "image_map", "image_pattern", "importance", "inside", "inside_vector", "int",
    "intensity_max", "interior", "interior_texture", "internal", "interpolate", "intersection", "intervals",
    "inv_pnormal_cure", "inverse", "ior", "ip_count", "ip_fade_distance", "ip_frequency", "ip_phase", "ip_scale",
    "ip_seed", "irid", "irid_wavelength", "isosurface", "it_depth", "it_frequency", "it_lambda", "it_octaves",
    "it_omega", "it_scale", "it_translate", "jitter", "jpeg", "julia", "julia_fractal", "lambda", "lathe", "lemon",
    "leopard", "light_group", "light_source", "linear_spline", "linear_sweep", "ln", "lo", "load_file", "location",
    "log", "log2", "look_at", "looks_like", "low_error_factor", "magnet", "major_radius", "mandel", "map_type",
    "marble", "material", "material_map", "matrix", "max", "max_extent", "max_gradient", "max_iteration", "max_sample",
    "max_trace", "max_trace_level", "maximum_reuse", "media", "media_attenuation", "media_interaction", "merge", "mesh",
    "mesh2", "mesh_camera", "metallic", "method", "metric", "micro", "min", "min_extent", "minimum_reuse", "mixed",
    "mm_per_unit", "mod", "mortar", "natural_spline", "nearest_count", "no", "no_bump_scale", "no_image",
    "no_radiosity", "no_reflection", "no_shadow", "noise_cubed", "noise_generator", "normal", "normal_indices",
    "normal_map", "normal_vectors", "now", "number_of_sides", "number_of_tiles", "object", "octaves", "off", "offset",
    "omega", "omnimax", "on", "once", "onion", "optional", "orient", "orientation", "orthographic", "ovus", "panoramic",
    "parallel", "parametric", "pass_through", "pattern", "pattern_modifiers", "pavement", "perspective", "pgm", "phase",
    "phong", "phong_odebla", "phong_size", "photons", "pi", "pigment", "pigment_map", "pigment_pattern", "planar",
    "plane", "png", "point_at", "poly", "poly_wave", "polygon", "polynomial", "pot", "potential", "pow", "ppm",
    "precision", "precompute", "premultiplied", "pretrace_end", "pretrace_start", "prism", "prod", "projected_through",
    "pwr", "quadratic_spline", "quadric", "quartic", "quaternion", "quick_color", "quick_colour", "quilted", "radial",
    "radians", "radiosity", "radius", "rainbow", "ramp_wave", "rand", "ratio", "raw_wave", "reciprocal",
    "recursion_limit", "red", "reflection", "refraction", "repeat", "report", "rgb", "rgbf", "rgbft", "rgbt", "right",
    "ripples", "rotate", "roughness", "samples", "save_file", "scale", "scallop_wave", "scattering", "seed", "seedfi",
    "select", "shadow_tolerance", "shadowless", "sin", "sine_wave", "sinh", "sint16be", "sint16le", "sint32be",
    "sint32le", "sint8", "size", "sky", "sky_sphere", "slice", "slope", "slope_map", "smooth", "smooth_triangle",
    "soft_object", "solid", "sor", "spacing", "specular", "specular_odebla", "sphere", "sphere_sweep", "spherical",
    "spiral1", "spiral2", "spline", "split_union", "spotlight", "sqr", "sqrt", "square", "srgb", "srgbf", "srgbft",
    "srgbt", "str", "strcmp", "strength", "strlen", "strlwr", "strupr", "sturm", "substr", "subsurface", "sum",
    "superellipsoid", "sys", "t", "tan", "tanh", "target", "tau", "text", "texture", "texture_list", "texture_map",
    "tga", "thickness", "threshold", "tiff", "tightness", "tiled", "tiling", "tolerance", "toroidal", "torus", "trace",
    "transform", "translate", "translucency", "transmit", "triangle", "triangle_wave", "triangular", "true", "ttf",
    "turbulence", "type", "u", "u_steps", "uint16be", "uint16le", "uint32be", "uint32le", "uint8", "ultra_wide_angle",
    "union", "unofficial", "up", "use_alpha", "use_color", "use_colour", "use_index", "user_defined", "utf8",
    "uv_indices", "uv_mapping", "uv_vectors", "v", "v_steps", "val", "variance", "vaxis_rotate", "vcross", "vdot",
    "version", "vertex_vectors", "vlength", "vnormalize", "vrotate", "vstr", "vturbulence", "warp", "water_level",
    "waves", "width", "wood", "word_get", "word_is", "wrinkles", "x", "xyz", "y", "yes", "z" };
static char oKeyWords[][26] { "agate_turb", "alpha", "ambient_light", "clock_delta", "composite", "evaluate",
    "final_clock", "final_frame", "fog_alt", "fog_offset", "fog_type", "frame_number", "hf_gray_16", "hf_grey_16",
    "initial_clock", "initial_frame", "max_intersections", "number_of_waves", "polarity", "prefix",
    "reflection_exponent", "sblue", "sgreen", "spotted", "sred", "suffix", "tile2", "tiles", "turb_depth" };
static char nKeyWords[][26] { "amplify", "bevy", "cmap", "e", "fork_str", "fork_val", "function_interval", "getenv",
    "global", "grooved", "hard_object", "hi", "id_type", "intensity_max", "inv_pnormal_cure", "ip_count",
    "ip_fade_distance", "ip_frequency", "ip_phase", "ip_scale", "ip_seed", "it_depth", "it_frequency", "it_lambda",
    "it_octaves", "it_omega", "it_scale", "it_translate", "lo", "log2", "micro", "noise_cubed", "pattern_modifiers",
    "phong_odebla", "raw_wave", "report", "seedfi", "shadow_tolerance", "soft_object", "specular_odebla", "tiled",
    "uint32be", "uint32le", "word_get", "word_is" };

static char iDirectiveWords[][26] { "!#", "#break", "#case", "#debug", "#declare", "#else", "#elseif",
    "#end", "#error", "#for", "#if", "#ifdef", "#ifndef", "#include", "#local", "#macro", "#open", "#range",
    "#read", "#switch", "#undef", "#warning", "#while", "#write", "break", "case", "debug", "declare", "else",
    "elseif", "end", "error", "for", "if", "ifdef", "ifndef", "include", "local", "macro", "open", "range", "read",
    "switch", "undef", "warning", "while", "write" };
static char oDirectiveWords[][26] { "!#", "#render", "#statistics", "render", "statistics" };
static char nDirectiveWords[][26] { "!#" };

static char iOddWords[][26] { "breakpoint", "dtag" };
static char oOddWords[][26] { "distance_maximum", "obj", "prefix", "suffix" };
static char nOddWords[][26] { "!#" };

// This a duplicate of one in parser.cpp optionally used there for no lower case ID checking.
static char iFnNames[][26] {"cf_abs", "cf_acos", "cf_acosh", "cf_add", "cf_asin", "cf_asinh", "cf_atan", "cf_atanh",
    "cf_cmplx", "cf_conj", "cf_cos", "cf_cosh", "cf_div", "cf_exp", "cf_imag", "cf_inv", "cf_ln", "cf_log", "cf_mul",
    "cf_neg", "cf_norm", "cf_phase", "cf_polar", "cf_pow", "cf_proj", "cf_real", "cf_sin", "cf_sinh", "cf_sqr",
    "cf_sqrt", "cf_sub", "cf_tan", "cf_tanh", "f_agate", "f_boom", "f_box", "f_boxb", "f_bump", "f_cbrt", "f_clock",
    "f_cone", "f_cubic_wave", "f_cylinder", "f_dec2x_f32", "f_dec3x_f21", "f_dec_nbit3c", "f_distribution",
    "f_dturbulence", "f_eblob", "f_ellipsoid", "f_elliptical_cylinder", "f_elliptical_sphrswp", "f_enc2x_f32",
    "f_enc3x_f21", "f_enc_nbit3c", "f_gradient", "f_granite23", "f_hash", "f_helix1", "f_helix2", "f_hessian_dist",
    "f_hetero_mf", "f_hypot", "f_interior_fade", "f_labsweep", "f_lame", "f_length", "f_length_sqr", "f_marble",
    "f_max1to8pairs", "f_min1to8pairs", "f_morph2to9fncts", "f_mult1to8pairs", "f_noise3d", "f_noise_cubed", "f_normal",
    "f_normalized", "f_npmod", "f_null", "f_odd", "f_onion", "f_paraboloid", "f_ph", "f_planar", "f_polyhedron",
    "f_quartic_cylinder", "f_quilt_cubic", "f_r", "f_radial", "f_radial_tubes", "f_randn", "f_range2range", "f_remap",
    "f_repeat", "f_revolution", "f_ridge", "f_ridged_mf", "f_rotate2d", "f_rounded_box", "f_scallop_wave",
    "f_scallop_wavef", "f_sign", "f_sign0", "f_signpow", "f_signsqrt", "f_sine_wave", "f_sine_wavef", "f_snoise3d",
    "f_sphere", "f_spiral", "f_superellipsoid", "f_supertorus", "f_th", "f_torus", "f_toupie", "f_triangle_wave",
    "f_turbulence", "f_upoly", "f_vangle", "f_wood", "f_wrap"};
static char oFnNames[][26] { "f_algbr_cyl1", "f_algbr_cyl2", "f_algbr_cyl3", "f_algbr_cyl4", "f_bicorn", "f_bifolia",
    "f_blob", "f_blob2", "f_boy_surface", "f_comma", "f_cross_ellipsoids", "f_crossed_trough", "f_cubic_saddle",
    "f_cushion", "f_devils_curve", "f_devils_curve_2d", "f_dupin_cyclid", "f_flange_cover", "f_folium_surface",
    "f_folium_surface_2d", "f_glob", "f_heart", "f_helical_torus", "f_hex_x", "f_hex_y", "f_hunt_surface",
    "f_hyperbolic_torus", "f_isect_ellipsoids", "f_kampyle_of_eudoxus", "f_kampyle_of_eudoxus_2d", "f_klein_bottle",
    "f_kummer_surface_v1", "f_kummer_surface_v2", "f_lemniscate_of_gerono", "f_lemniscate_of_gerono_2d", "f_mesh1",
    "f_mitre", "f_nodal_cubic", "f_noise_generator", "f_ovals_of_cassini", "f_parabolic_torus", "f_pillow",
    "f_piriform", "f_piriform_2d", "f_poly4", "f_polytubes", "f_quantum", "f_quartic_paraboloid", "f_quartic_saddle",
    "f_spikes", "f_spikes_2d", "f_steiners_roman", "f_strophoid", "f_strophoid_2d", "f_torus2", "f_torus_gumdrop",
    "f_umbrella", "f_witch_of_agnesi", "f_witch_of_agnesi_2d" };
static char nFnNames[][26] { "cf_abs", "cf_acos", "cf_acosh", "cf_add", "cf_asin", "cf_asinh", "cf_atan", "cf_atanh",
    "cf_cmplx", "cf_conj", "cf_cos", "cf_cosh", "cf_div", "cf_exp", "cf_imag", "cf_inv", "cf_ln", "cf_log", "cf_mul",
    "cf_neg", "cf_norm", "cf_phase", "cf_polar", "cf_pow", "cf_proj", "cf_real", "cf_sin", "cf_sinh", "cf_sqr",
    "cf_sqrt", "cf_sub", "cf_tan", "cf_tanh", "f_agate", "f_boom", "f_box", "f_boxb", "f_bump", "f_cbrt", "f_clock",
    "f_cone", "f_cubic_wave", "f_cylinder", "f_dec2x_f32", "f_dec3x_f21", "f_dec_nbit3c", "f_distribution",
    "f_dturbulence", "f_eblob", "f_elliptical_cylinder", "f_elliptical_sphrswp", "f_enc2x_f32", "f_enc3x_f21",
    "f_enc_nbit3c", "f_gradient", "f_granite23", "f_hash", "f_hessian_dist", "f_hypot", "f_interior_fade", "f_labsweep",
    "f_lame", "f_length", "f_length_sqr", "f_marble", "f_max1to8pairs", "f_min1to8pairs", "f_morph2to9fncts",
    "f_mult1to8pairs", "f_noise_cubed", "f_normal", "f_normalized", "f_npmod", "f_null", "f_onion", "f_planar",
    "f_polyhedron", "f_quilt_cubic", "f_radial", "f_radial_tubes", "f_randn", "f_range2range", "f_remap", "f_repeat",
    "f_revolution", "f_rotate2d", "f_scallop_wave", "f_scallop_wavef", "f_sign", "f_sign0", "f_signpow", "f_signsqrt",
    "f_sine_wave", "f_sine_wavef", "f_snoise3d", "f_supertorus", "f_toupie", "f_triangle_wave", "f_turbulence",
    "f_upoly", "f_vangle", "f_wood", "f_wrap" };

static char iFakeByFnctWords[][26] { "!#" }; // None in povr. Some old now defined mixed case.
static char oFakeByFnctWords[][26] { "acosd", "adj_range", "adj_range2", "asind", "atan2d", "atand", "clamp", "clip",
    "cosd", "even", "f_boxed", "f_bozo", "f_brick", "f_bumps", "f_checker", "f_crackle", "f_cylindrical", "f_dents",
    "f_gradientX", "f_gradientY", "f_gradientZ", "f_granite", "f_hexagon", "f_leopard", "f_mandel", "f_ripples",
    "f_spherical", "f_spiral1", "f_spiral2", "f_spotted", "f_sqr", "f_waves", "f_wrinkles", "max3", "min3", "odd",
    "sgn", "sind", "tand" };
static char nFakeByFnctWords[][26] { "!#" };

static char iOptionNames[][26] { "+a", "+ac", "+ag", "+am", "+b", "+bm", "+bs", "+c", "+cc", "+d", "+ec", "+ef", "+ef0",
    "+ep", "+er", "+f", "+ga", "+gd", "+gf", "+gi", "+gp", "+gr", "+gs", "+gw", "+h", "+hi", "+hr", "+i", "+j", "+k",
    "+kc", "+kf", "+kff", "+kfi", "+ki", "+kla", "+l", "+mb", "+mi", "+mv", "+o", "+p", "+q", "+r", "+rf", "+rfi",
    "+rfo", "+rp", "+rs", "+rtc", "+rtr", "+rvp", "+sc", "+sf", "+sf0", "+sp", "+sr", "+ss", "+stp", "+su", "+th",
    "+ua", "+ur", "+v", "+w", "+wl", "+wt", "all_console", "all_file", "antialias", "antialias_confidence",
    "antialias_depth", "antialias_gamma", "antialias_threshold", "append_file", "bits_per_color", "bits_per_colour",
    "bounding", "bounding_method", "bounding_threshold", "bsp_baseaccesscost", "bsp_childaccesscost", "bsp_isectcost",
    "bsp_maxdepth", "bsp_misschance", "clock", "clockless_animation", "compression", "continue_trace",
    "create_continue_trace_log", "create_ini", "cyclic_animation", "debug_console", "debug_file", "declare", "display",
    "display_gamma", "dither", "dither_method", "end_column", "end_row", "fatal_console", "fatal_error_command",
    "fatal_error_return", "fatal_file", "file_gamma", "final_clock", "final_frame", "frame_step", "grayscale_output",
    "greyscale_output", "height", "high_reproducibility", "include_header", "include_ini", "initial_clock",
    "initial_frame", "input_file_name", "jitter", "jitter_amount", "library_path", "max_image_buffer_memory",
    "output_alpha", "output_file_name", "output_file_type", "output_to_file", "pause_when_done", "post_frame_command",
    "post_frame_return", "post_scene_command", "post_scene_return", "pre_frame_command", "pre_frame_return",
    "pre_scene_command", "pre_scene_return", "preview_end_size", "preview_start_size", "quality", "radiosity_file_name",
    "radiosity_from_file", "radiosity_to_file", "radiosity_vain_pretrace", "real_time_capture", "real_time_raytracing",
    "remove_bounds", "render_block_size", "render_block_step", "render_console", "render_file", "render_pattern",
    "sampling_method", "split_unions", "start_column", "start_row", "statistic_console", "statistic_file",
    "stochastic_seed", "subset_end_frame", "subset_start_frame", "user_abort_command", "user_abort_return", "verbose",
    "version", "warning_console", "warning_file", "warning_level", "width", "work_threads" };
static char oOptionNames[][26] { "+a0", "+hn", "+hs", "+ht", "+qr", "+ud", "+uf", "+ul", "+uo", "+uv", "+x",
    "buffer_output", "buffer_size", "create_histogram", "draw_vistas", "field_render", "histogram_grid_size",
    "histogram_name", "histogram_type", "light_buffer", "odd_field", "palette", "radiosity", "test_abort",
    "test_abort_count", "video_mode", "vista_buffer" };
static char nOptionNames[][26] { "+rtc", "real_time_capture" };

//******************************************************************************

char *Parser::Parse_C_String(bool pathname)
{
    UCS2 *str = Parse_String(pathname);
    char *New = UCS2_To_String(str);

    POV_FREE(str);

    return New;
}

void Parser::ParseString(UTF8String& s, bool pathname)
{
    /// @todo Add support for non-ASCII strings.

    UCS2 *str = Parse_String(pathname);
    char *New = UCS2_To_String(str);

    s = New;

    POV_FREE(str);
}

//******************************************************************************

UCS2 *Parser::Parse_String(bool pathname, bool require)
{
    UCS2 *New = nullptr;
    int len = 0;
    const UCS2String* pString;
    const StringValue* stringValue = nullptr;

    EXPECT
        CASE(STRING_LITERAL_TOKEN)
            /// @todo Add back support for non-ASCII string encodings.
            stringValue = dynamic_cast<const StringValue*>(mToken.raw.value.get());
            POV_PARSER_ASSERT(stringValue != nullptr);

            if (pathname)
            {
                // Historically, escape sequences were ignored when parsing for a filename.
                // As of POV-Ray v3.8, this has been changed.

                if (sceneData->EffectiveLanguageVersion() >= 380)
                {
                    if (stringValue->IsAmbiguous())
                    {
#if POV_BACKSLASH_IS_PATH_SEPARATOR
                        Warning("Backslash encountered while parsing for a filename."
                                " As of POV-Ray v3.8, this is interpreted as an escape sequence just like in"
                                " any other string literal. If this is supposed to be a path separator,"
                                " use a forward slash instead.");
#endif

                    }
                    pString = &stringValue->GetData();
                }
                else
                {
                    if (stringValue->IsAmbiguous())
                    {
#if POV_BACKSLASH_IS_PATH_SEPARATOR
                        Warning("Backslash encountered while parsing for a filename."
                                " In legacy (pre-v3.8) scenes, this is NOT interpreted as the start of an"
                                " escape sequence. However, for future compatibility it is recommended to"
                                " use a forward slash as path separator instead.");
#else
                        Warning("Backslash encountered while parsing for a filename."
                                " In legacy (pre-v3.8) scenes, this is NOT interpreted as the start of an"
                                " escape sequence.");
#endif
                    }
                    pString = &stringValue->GetFileName();
                }
            }
            else
            {
                pString = &stringValue->GetData();
            }

            len = pString->size() + 1;
            New = reinterpret_cast<UCS2 *>(POV_MALLOC(len * sizeof(UCS2), "UCS2 String"));
            std::memcpy(reinterpret_cast<void *>(New),
                        reinterpret_cast<const void *>(pString->c_str()),
                        len * sizeof(UCS2));
            EXIT
        END_CASE

        CASE(STR_TOKEN)
            New = Parse_Str(pathname);
            EXIT
        END_CASE

        CASE(VSTR_TOKEN)
            New = Parse_VStr(pathname);
            EXIT
        END_CASE

        CASE(CONCAT_TOKEN)
            New = Parse_Concat(pathname);
            EXIT
        END_CASE

        CASE(CHR_TOKEN)
            New = Parse_Chr(pathname);
            EXIT
        END_CASE

        CASE(DATETIME_TOKEN)
            New = Parse_Datetime(pathname);
            EXIT
        END_CASE

        CASE(GETENV_TOKEN)
            New = Parse_Getenv(pathname);
            EXIT
        END_CASE

        CASE(FORK_STR_TOKEN)
            New = Parse_ForkStr((int)Parse_Float_Param());
            EXIT
        END_CASE

        CASE(SUBSTR_TOKEN)
            New = Parse_Substr(pathname);
            EXIT
        END_CASE

        CASE(STRUPR_TOKEN)
            New = Parse_Strupr(pathname);
            EXIT
        END_CASE

        CASE(STRLWR_TOKEN)
            New = Parse_Strlwr(pathname);
            EXIT
        END_CASE

        CASE(STRING_ID_TOKEN)
            len = UCS2_strlen(CurrentTokenDataPtr<UCS2*>()) + 1;
            New = reinterpret_cast<UCS2 *>(POV_MALLOC(len * sizeof(UCS2), "UCS2 String"));
            std::memcpy(reinterpret_cast<void *>(New), CurrentTokenDataPtr<void*>(), len * sizeof(UCS2));
            EXIT
        END_CASE

        CASE(WORD_GET_TOKEN)
            New = Parse_WordGet(pathname);
            EXIT
        END_CASE

        OTHERWISE
            if (require)
                Expectation_Error("string expression");
            else
            {
                UNGET
                EXIT
            }
        END_CASE
    END_EXPECT

    return New;
}

//******************************************************************************

std::string Parser::Parse_SysString(bool pathname, bool require)
{
    UCS2 *cstr = Parse_String(pathname, require);
    std::string ret(UCS2toSysString(cstr));
    POV_FREE(cstr);
    return ret;
}

//******************************************************************************

UCS2 *Parser::Parse_Str(bool pathname)
{
    char *p;
    char temp3[128];
    char temp4[256];
    double val;
    int l, d;

    Parse_Paren_Begin();
    val = Parse_Float();
    Parse_Comma();
    l = (int)Parse_Float();
    Parse_Comma();
    d = (int)Parse_Float();
    Parse_Paren_End();

    p = temp3;
    *p++ = '%';
    if (l > 0)
    {
        p += sprintf(p, "%d", l);
    }
    else
    {
        if (l)
            p += sprintf(p, "0%d", abs(l));
    }

    if (d >= 0)
        p += sprintf(p, ".%d", d);
    strcpy(p, "f");

    // a very large floating point value (e.g. 1e251) will overflow the buffer.
    // TODO: consider changing to %g rather than %f for large numbers (e.g.
    // anything over 1e+64 for example). for now, we will only use %g if the
    // snprintf filled the buffer.
    // NB `snprintf` may report errors via a negative return value.
    if (((l = std::snprintf(temp4, sizeof(temp4), temp3, val)) >= sizeof(temp4)) || (l < 0))
    {
        *p = 'g';

        // it should not be possible to overflow with %g. but just in case ...
        if (((l = std::snprintf(temp4, sizeof(temp4), temp3, val)) >= sizeof(temp4)) || (l < 0))
            strcpy(temp4, "<invalid>");
    }

    return String_To_UCS2(temp4);
}

//******************************************************************************

UCS2 *Parser::Parse_VStr(bool pathname)
{
    char *p;
    char temp3[128];
    char temp4[768];
    int l, d, vl;
    EXPRESS Express;
    int Terms;
    int Dim = 5;
    UCS2 *str;
    UCS2 *str2;
    UCS2 *New;

    Parse_Paren_Begin();

    vl = (int)Parse_Float();
    Parse_Comma();

    if (vl < 2)
        vl = 2;
    else if (vl > 5)
        vl = 5;
    Dim = vl;

    Terms = Parse_Unknown_Vector(Express);

    Parse_Comma();
    str = Parse_String(pathname);
    Parse_Comma();
    l = (int)Parse_Float();
    Parse_Comma();
    d = (int)Parse_Float();

    Parse_Paren_End();

    p = temp3;
    *(p++) = '%';
    if (l > 0)
    {
        sprintf(p, "%d", l);
        while (*p != '\0')
            p++;
    }
    else
    {
        if (l)
        {
            sprintf(p, "0%d", abs(l));
            while (*p != '\0')
                p++;
        }
    }

    if (d >= 0)
    {
        *(p++) = '.';
        sprintf(p, "%d", d);
        while (*(++p))
            ;
    }
    *(p++) = 'f';
    *p = '\0';

    sprintf(temp4, temp3, Express[X]);
    New = String_To_UCS2(temp4);       // add first component

    for (Terms = 1; Terms < Dim; Terms++)
    {
        New = UCS2_strcat(New, str);   // add separator
        sprintf(temp4, temp3, Express[Terms]);
        str2 = String_To_UCS2(temp4);
        New = UCS2_strcat(New, str2);  // add component
        POV_FREE(str2);
    }

    POV_FREE(str);

    return New;
}

//******************************************************************************

UCS2 *Parser::Parse_Concat(bool pathname)
{
    UCS2 *str;
    UCS2 *New;

    Parse_Paren_Begin();

    New = Parse_String();

    EXPECT
        CASE(RIGHT_PAREN_TOKEN)
            UNGET
            EXIT
        END_CASE

        OTHERWISE
            UNGET
            Parse_Comma();
            str = Parse_String(pathname);
            New = UCS2_strcat(New, str);
            POV_FREE(str);
        END_CASE
    END_EXPECT

    Parse_Paren_End();

    return New;
}

//******************************************************************************

UCS2 *Parser::Parse_Chr(bool /*pathname*/)
{
    UCS2 *New;
    int d;

    New = reinterpret_cast<UCS2 *>(POV_MALLOC(sizeof(UCS2) * 2, "temporary string"));

    d = (int)Parse_Float_Param();
    if ((d < 0) || (d > 65535))
        Error("Value %d cannot be used in chr(...).", d);

    New[0] = d;
    New[1] = 0;

    return New;
}

//******************************************************************************

#define PARSE_NOW_VAL_LENGTH 200

UCS2 *Parser::Parse_Datetime(bool pathname)
{
    char *FormatStr = nullptr;
    bool CallFree;
    int vlen = 0;
    char val[PARSE_NOW_VAL_LENGTH + 1]; // Arbitrary size, usually a date format string is far less

    Parse_Paren_Begin();

    double FractionalDays = Parse_Float();

  //std::time_t timestamp = floor((Parse_Float() + (365*30+7)) * 24*60*60 + 0.5);
    std::time_t timestamp = floor((FractionalDays + (365*30+7)) * 24*60*60 + 0.5);

    Parse_Comma();
    EXPECT_ONE
        CASE(RIGHT_PAREN_TOKEN)
            UNGET
            CallFree = false;

            // Old note on why POV-Ray attempted use of 'Z': we use GMT as some
            // platforms (e.g. windows) have different ideas of what to print
            // when handling '%z'.

          //FormatStr = (char *)"%Y-%m-%d %H:%M:%S %Z"; // Long a bug in spec(%SZ). %Z is a GMT timezone indication.
            FormatStr = (char *)"%Y-%m-%d %H:%M:%S %z"; // %z offset from UTC in the ISO 8601 format (e.g. -0400)
        END_CASE

        OTHERWISE
            UNGET
            CallFree = true;
            FormatStr = Parse_C_String(pathname);
            if (FormatStr[0] == '\0')
            {
                POV_FREE(FormatStr);
                Error("Empty format string.");
            }
            if (strlen(FormatStr) > PARSE_NOW_VAL_LENGTH)
            {
                POV_FREE(FormatStr);
                Error("Format string too long.");
            }
        END_CASE
    END_EXPECT

    Parse_Paren_End();

    // NB don't wrap only the call to strftime() in the try, because visual C++ will, in release mode,
    // optimize the try/catch away since it doesn't believe that the RTL can throw exceptions. since
    // the windows version of POV hooks the invalid parameter handler RTL callback and throws an exception
    // if it's called, they can.
    try
    {
        std::tm t = boost::posix_time::to_tm(boost::posix_time::from_time_t(timestamp));

#if (1) //      This current povr version... Could use some clean up as things settle.
        // If users want the time zone clock time no matter whether it being daylight savings
        // or not, they can specify a negative day value of -100000 (-1e5) or less. One can only
        // move backward to a max 1970 (1970-01-01 00:00:00) in any case so the value is good
        // for more than 200 years as of 2021.
        std::time_t tt;
        if (FractionalDays <= -1e5)
        {
            tt = std::time(nullptr);
        }
        else
        {
            tt = std::mktime(&t); // This really just to copy to tt - best can tell.
        }

        // TODO FIXME - we should either have this locale setting globally, or
        // avoid it completely; in either case it shouldn't be *here*. Whether
        // the setting is useful at all depends on the user specifying a format
        // string to strftime() creating a meaningful difference in output for
        // the given locale. With the simple defaults this often not the case.

        setlocale(LC_TIME,"");  // Get the local preferred format
     // vlen = strftime(val, PARSE_NOW_VAL_LENGTH, FormatStr, &t); // Changing only %z stuff relative to dst, not the time.
        vlen = strftime(val, PARSE_NOW_VAL_LENGTH, FormatStr, std::localtime(&tt));  // Used to enable local clock time to current dst
        setlocale(LC_TIME,"C"); // Being paranoid here. Likely doesn't matter, but "C" is what the C++ run time should run at start.
#else
        setlocale(LC_TIME,""); // Get the local preferred format    // This original v3.8 code
// std::time_t tt = std::mktime(&t);
// vlen = strftime(val, PARSE_NOW_VAL_LENGTH, FormatStr, std::localtime(&tt));
        vlen = strftime(val, PARSE_NOW_VAL_LENGTH, FormatStr, &t);
#endif
    }
    catch (pov_base::Exception& e)
    {
        // the windows version of strftime calls the invalid parameter handler if
        // it gets a bad format string. this will in turn raise an exception of type
        // kParamErr. if the exception isn't that, allow normal exception processing
        // to continue, otherwise we issue a more useful error message.
        if ((e.codevalid() == false) || (e.code() != kParamErr))
            throw;
        vlen = 0;
    }
    if (vlen == PARSE_NOW_VAL_LENGTH) // on error: max for libc 4.4.1 & before
        vlen = 0; // return an empty string on error (content of val[] is undefined)
    val[vlen]='\0'; // whatever, that operation is now safe (and superfluous except for error)

    if (CallFree)
    {
        POV_FREE(FormatStr);
    }

    if (vlen == 0)
        Error("Invalid formatting code in format string, or resulting string too long.");

    return String_To_UCS2(val);
}

//******************************************************************************

UCS2 *Parser::Parse_Getenv(bool pathname)
{
    char *Str = nullptr;
    const char* envPtr;

    Parse_Paren_Begin();

    EXPECT_ONE
        CASE(RIGHT_PAREN_TOKEN)
            Expectation_Error("string expression");
            UNGET
        END_CASE

        OTHERWISE
            UNGET
            Str = Parse_C_String(pathname); // Limited to <256 bytes
            if (Str[0] == '\0')
            {
                POV_FREE(Str);
                Error("Empty environment variable name string.");
            }
            if (strlen(Str) > 201)
            {
                POV_FREE(Str);
                Error("Environment variable name string is too long (>200 bytes).");
            }
        END_CASE
    END_EXPECT

    if (envPtr = std::getenv(Str))
    {
        POV_FREE(Str);
    }
    else
    {
        POV_FREE(Str);
        Error("Environment variable name is not set.");
    }

    Parse_Paren_End();

    return String_To_UCS2(envPtr);
}

//******************************************************************************

UCS2 *Parser::Parse_ForkStr(int index)
{
    if ((index >= FORK_STR_FORK_VAL_MAX_COUNT) || (index < 0))
        Error("The fork_str(n), n index value must be one of [0..%d].",
                FORK_STR_FORK_VAL_MAX_COUNT-1);

    return String_To_UCS2(Fork_Strings[index]); // Fork_Strings static char array at top.
}

//******************************************************************************

UCS2 *Parser::Parse_WordGet(bool pathname)
{
    char *Str = nullptr;
    const char* envPtr;
    int index, count;
    char *w;

    // -6 +6
    int CntoOptionNames = sizeof(oOptionNames)/sizeof(oOptionNames[0]);
    int CntiOptionNames = sizeof(iOptionNames)/sizeof(iOptionNames[0]);
    int CntnOptionNames = sizeof(nOptionNames)/sizeof(nOptionNames[0]);
    // -5 +5
    int CntoFakeByFnctWords = sizeof(oFakeByFnctWords)/sizeof(oFakeByFnctWords[0]);
    int CntiFakeByFnctWords = sizeof(iFakeByFnctWords)/sizeof(iFakeByFnctWords[0]);
    int CntnFakeByFnctWords = sizeof(nFakeByFnctWords)/sizeof(nFakeByFnctWords[0]);
    // -4 +4
    int CntoFnNames = sizeof(oFnNames)/sizeof(oFnNames[0]);
    int CntiFnNames = sizeof(iFnNames)/sizeof(iFnNames[0]);
    int CntnFnNames = sizeof(nFnNames)/sizeof(nFnNames[0]);
    // -3 +3
    int CntoOddWords = sizeof(oOddWords)/sizeof(oOddWords[0]);
    int CntiOddWords = sizeof(iOddWords)/sizeof(iOddWords[0]);
    int CntnOddWords = sizeof(nOddWords)/sizeof(nOddWords[0]);
    // -2 +2
    int CntoDirectiveWords = sizeof(oDirectiveWords)/sizeof(oDirectiveWords[0]);
    int CntiDirectiveWords = sizeof(iDirectiveWords)/sizeof(iDirectiveWords[0]);
    int CntnDirectiveWords = sizeof(nDirectiveWords)/sizeof(nDirectiveWords[0]);
    // -1 +1
    int CntoKeyWords = sizeof(oKeyWords)/sizeof(oKeyWords[0]);
    int CntiKeyWords = sizeof(iKeyWords)/sizeof(iKeyWords[0]);
    int CntnKeyWords = sizeof(nKeyWords)/sizeof(nKeyWords[0]);

    Parse_Paren_Begin();

    EXPECT_ONE
        CASE(RIGHT_PAREN_TOKEN)
            Expectation_Error("string expression");
            UNGET
        END_CASE

        OTHERWISE
            UNGET
            Str = Parse_C_String(pathname); // Limited to <256 bytes
            if (Str[0] == '\0')
            {
                POV_FREE(Str);
                Error("Empty word array name name string.");
            }
            if (strlen(Str) > 201)
            {
                POV_FREE(Str);
                Error("Word array name string is too long (>200 bytes).");
            }
        END_CASE
    END_EXPECT

    Parse_Comma();
    index = (int)Parse_Float();

    count = 0;
    if ( Str[0]=='0' )
    {   // -6 +6
        if (strcmp(Str,"0oOptionNames")==0) { count = CntoOptionNames; w=&oOptionNames[0][0]; }
        if (strcmp(Str,"0iOptionNames")==0) { count = CntiOptionNames; w=&iOptionNames[0][0]; }
        if (strcmp(Str,"0nOptionNames")==0) { count = CntnOptionNames; w=&nOptionNames[0][0]; }
        // -5 +5
        if (strcmp(Str,"0oFakeByFnctWords")==0) { count = CntoFakeByFnctWords; w=&oFakeByFnctWords[0][0]; }
        if (strcmp(Str,"0iFakeByFnctWords")==0) { count = CntiFakeByFnctWords; w=&iFakeByFnctWords[0][0]; }
        if (strcmp(Str,"0nFakeByFnctWords")==0) { count = CntnFakeByFnctWords; w=&nFakeByFnctWords[0][0]; }
        // -4 +4
        if (strcmp(Str,"0oFnNames")==0) { count = CntoFnNames; w=&oFnNames[0][0]; }
        if (strcmp(Str,"0iFnNames")==0) { count = CntiFnNames; w=&iFnNames[0][0]; }
        if (strcmp(Str,"0nFnNames")==0) { count = CntnFnNames; w=&nFnNames[0][0]; }
        // -3 +3
        if (strcmp(Str,"0oOddWords")==0) { count = CntoOddWords; w=&oOddWords[0][0]; }
        if (strcmp(Str,"0iOddWords")==0) { count = CntiOddWords; w=&iOddWords[0][0]; }
        if (strcmp(Str,"0nOddWords")==0) { count = CntnOddWords; w=&nOddWords[0][0]; }
        // -2 +2
        if (strcmp(Str,"0oDirectiveWords")==0) { count = CntoDirectiveWords; w=&oDirectiveWords[0][0]; }
        if (strcmp(Str,"0iDirectiveWords")==0) { count = CntiDirectiveWords; w=&iDirectiveWords[0][0]; }
        if (strcmp(Str,"0nDirectiveWords")==0) { count = CntnDirectiveWords; w=&nDirectiveWords[0][0]; }
        // -1 +1
        if (strcmp(Str,"0oKeyWords")==0) { count = CntoKeyWords; w=&oKeyWords[0][0]; }
        if (strcmp(Str,"0iKeyWords")==0) { count = CntiKeyWords; w=&iKeyWords[0][0]; }
        if (strcmp(Str,"0nKeyWords")==0) { count = CntnKeyWords; w=&nKeyWords[0][0]; }

        if (count <= 0)
        {
            POV_FREE(Str);
            Error("Word array name string was not recognized. Try test with '0iKeyWords'.");
        }

        if ((index < 0) || (index >= count))
        {
            char ErrStr[512];
            snprintf(ErrStr,sizeof(ErrStr),"Passed index %d not in array %s range of [0..%d].\n",
                    index,Str,count);
            POV_FREE(Str);
            Error(ErrStr);
        }

        w += index*26;
    }
    else
    {
        POV_FREE(Str);
        Error("Word array name string must be prefixed by 0 as in '0iKeyWords'.");
    }

    POV_FREE(Str);

    Parse_Paren_End();

    return String_To_UCS2(w);
}

//******************************************************************************

UCS2 *Parser::Parse_Substr(bool pathname)
{
    UCS2 *str;
    UCS2 *New;
    int l, d;

    Parse_Paren_Begin();

    str = Parse_String(pathname);
    Parse_Comma();
    l = (int)Parse_Float();
    Parse_Comma();
    d = (int)Parse_Float();

    Parse_Paren_End();

    if (((l + d - 1) > UCS2_strlen(str)) || (l < 0) || (d < 0))
        Error("Illegal parameters in substr.");

    New = reinterpret_cast<UCS2 *>(POV_MALLOC(sizeof(UCS2) * (d + 1), "temporary string"));
    UCS2_strncpy(New, &(str[l - 1]), d);
    New[d] = 0;

    POV_FREE(str);

    return New;
}

//******************************************************************************

UCS2 *Parser::Parse_Strupr(bool pathname)
{
    UCS2 *New;

    Parse_Paren_Begin();

    New = Parse_String(pathname);
    UCS2_strupr(New);

    Parse_Paren_End();

    return New;
}

//******************************************************************************

UCS2 *Parser::Parse_Strlwr(bool pathname)
{
    UCS2 *New;

    Parse_Paren_Begin();

    New = Parse_String(pathname);
    UCS2_strlwr(New);

    Parse_Paren_End();

    return New;
}

//******************************************************************************

UCS2 *Parser::String_To_UCS2(const char *str)
{
    UCS2 *char_string = nullptr;
    UCS2 *char_array = nullptr;
    int char_array_size = 0;
    int utf8arraysize = 0;
    unsigned char *utf8array = nullptr;
    int index_in = 0;
    int index_out = 0;
    char *dummy_ptr = nullptr;
    int i = 0;

    if (strlen(str) == 0)
    {
        char_string = reinterpret_cast<UCS2 *>(POV_MALLOC(sizeof(UCS2), "UCS2 String"));
        char_string[0] = 0;

        return char_string;
    }

    char_array_size = (int)strlen(str);
    char_array = reinterpret_cast<UCS2 *>(POV_MALLOC(char_array_size * sizeof(UCS2), "Character Array"));
    for (i = 0; i < char_array_size; i++)
    {
        if (sceneData->EffectiveLanguageVersion() < 350)
            char_array[i] = (unsigned char)(str[i]);
        else
        {
            char_array[i] = str[i] & 0x007F;
            if (char_array[i] != str[i])
            {
                char_array[i] = ' ';
                PossibleError("Unexpected non-ASCII character has been replaced by space character.");
            }
        }
    }

    char_string = reinterpret_cast<UCS2 *>(POV_MALLOC((char_array_size + 1) * sizeof(UCS2), "UCS2 String"));
    for (index_in = 0, index_out = 0; index_in < char_array_size; index_in++, index_out++)
        char_string[index_out] = char_array[index_in];

    char_string[index_out] = 0;
    index_out++;

    if (char_array != nullptr)
        POV_FREE(char_array);

    return char_string;
}

//******************************************************************************

UCS2 *Parser::String_Literal_To_UCS2(const std::string& str)
{
    UCS2 *char_string = nullptr;
    UCS2 *char_array = nullptr;
    std::string::size_type char_array_size = 0;
    int utf8arraysize = 0;
    unsigned char *utf8array = nullptr;
    int index_in = 0;
    int index_out = 0;
    char buffer[8];
    char *dummy_ptr = nullptr;
    int i = 0;

    if (str.length() == 0)
    {
        char_string = reinterpret_cast<UCS2 *>(POV_MALLOC(sizeof(UCS2), "UCS2 String"));
        char_string[0] = 0;

        return char_string;
    }

    char_array_size = str.length();
    char_array = reinterpret_cast<UCS2 *>(POV_MALLOC(char_array_size * sizeof(UCS2), "Character Array"));
    for (i = 0; i < char_array_size; i++)
    {
        if (sceneData->EffectiveLanguageVersion() < 350)
            char_array[i] = (unsigned char)(str[i]);
        else
        {
            char_array[i] = str[i] & 0x007F;
            if (char_array[i] != str[i])
            {
                char_array[i] = ' ';
                PossibleError("Unexpected non-ASCII character has been replaced by space character.");
            }
        }
    }

    char_string = reinterpret_cast<UCS2 *>(POV_MALLOC((char_array_size + 1) * sizeof(UCS2), "UCS2 String"));
    for (index_in = 0, index_out = 0; index_in < char_array_size; index_in++, index_out++)
    {
        if (char_array[index_in] == '\\')
        {
            index_in++;

            switch (char_array[index_in])
            {
                case 'a':
                    char_string[index_out] = 0x07;
                    break;
                case 'b':
                    char_string[index_out] = 0x08;
                    break;
                case 'f':
                    char_string[index_out] = 0x0c;
                    break;
                case 'n':
                    char_string[index_out] = 0x0a;
                    break;
                case 'r':
                    char_string[index_out] = 0x0d;
                    break;
                case 't':
                    char_string[index_out] = 0x09;
                    break;
                case 'v':
                    char_string[index_out] = 0x0b;
                    break;
                case '\0':
                    // [CLi] shouldn't happen, as having a backslash as the last character of a string literal would invalidate the string terminator
                    Error("Unexpected end of escape sequence in text string.");
                    break;
                case '\'':
                case '\"':
                case '\\':
                    char_string[index_out] = char_array[index_in];
                    break;
                case 'u':
                    if (index_in + 4 >= char_array_size)
                        Error("Unexpected end of escape sequence in text string.");

                    buffer[0] = char_array[++index_in];
                    buffer[1] = char_array[++index_in];
                    buffer[2] = char_array[++index_in];
                    buffer[3] = char_array[++index_in];
                    buffer[4] = 0;

                    char_string[index_out] = (UCS2)std::strtoul(buffer, &dummy_ptr, 16);
                    break;
                default:
                    char_string[index_out] = char_array[index_in];
                    POV_FREE(char_array);
                    char_array = nullptr;
                    Error( "Illegal escape sequence in string." );
                    break;
            }
        }
        else
            char_string[index_out] = char_array[index_in];
    }

    char_string[index_out] = 0;
    index_out++;

    char_string = reinterpret_cast<UCS2 *>(POV_REALLOC(char_string, index_out * sizeof(UCS2), "UCS2 String"));

    if (char_array != nullptr)
        POV_FREE(char_array);

    return char_string;
}

//******************************************************************************

char *Parser::UCS2_To_String(const UCS2 *str)
{
    char *str_out;
    char *strp;

    str_out = reinterpret_cast<char *>(POV_MALLOC(UCS2_strlen(str)+1, "C String"));

    for (strp = str_out; *str != 0; str++, strp++)
    {
        if ((*str > 127) && (sceneData->EffectiveLanguageVersion() >= 350))
            *strp = ' ';
        else
            *strp = (char)(*str);
    }

    *strp = 0;

    return str_out;
}


/*****************************************************************************
*
* FUNCTION
*
*   Convert_UTF8_To_UCS2
*
* INPUT
*
*   Array of bytes, length of this sequence
*
* OUTPUT
*
*   Size of the array of UCS2s returned
*
* RETURNS
*
*   Array of UCS2s (allocated with POV_MALLOC)
*
* AUTHOR
*
* DESCRIPTION
*
*   Converts UTF8 to UCS2 characters, however all surrogates are dropped.
*
* CHANGES
*
*   -
*
******************************************************************************/

UCS2 *Parser::Convert_UTF8_To_UCS2(const unsigned char *text_array, int *char_array_size)
{
    POV_PARSER_ASSERT(text_array);
    POV_PARSER_ASSERT(char_array_size);

    UCS2String s = UTF8toUCS2String(UTF8String(reinterpret_cast<const char*>(text_array)));
    UCS2String::size_type len = s.length();
    *char_array_size = len;

    if (len == 0)
        return nullptr;

    size_t size = (len+1)*sizeof(UCS2);

    UCS2 *char_array = reinterpret_cast<UCS2 *>(POV_MALLOC(size, "Character Array"));
    if (char_array == nullptr)
        throw POV_EXCEPTION_CODE(kOutOfMemoryErr);

    memcpy(char_array, s.c_str(), size);

    return char_array;
}

//******************************************************************************

UCS2 *Parser::UCS2_strcat(UCS2 *s1, const UCS2 *s2)
{
    int l1, l2;

    l1 = UCS2_strlen(s1);
    l2 = UCS2_strlen(s2);

    s1 = reinterpret_cast<UCS2 *>(POV_REALLOC(s1, sizeof(UCS2) * (l1 + l2 + 1), "UCS2 String"));

    UCS2_strcpy(&s1[l1], s2);

    return s1;
}

//******************************************************************************

void Parser::UCS2_strcpy(UCS2 *s1, const UCS2 *s2)
{
    for (; *s2 != 0; s1++, s2++)
        *s1 = *s2;

    *s1 = 0;
}

//******************************************************************************

void Parser::UCS2_strncpy(UCS2 *s1, const UCS2 *s2, int n)
{
    for (; (*s2 != 0) && (n > 0); s1++, s2++, n--)
        *s1 = *s2;

    *s1 = 0;
}

//******************************************************************************

void Parser::UCS2_strupr(UCS2 *str)
{
    bool err = false;

    while (true)
    {
        if (((int) *str < 0) || (*str > 127))
            err = true;
        else if (*str == 0)
            break;

        *str = toupper(*str);
        str++;
    }

    if (err == true)
        Warning("Non-ASCII character in string, strupr may not work as expected.");
}

//******************************************************************************

void Parser::UCS2_strlwr(UCS2 *str)
{
    bool err = false;

    while (true)
    {
        if (((int) *str < 0) || (*str > 127))
            err = true;
        else if (*str == 0)
            break;

        *str = tolower(*str);
        str++;
    }

    if (err == true)
        Warning("Non-ASCII character in string, strlwr may not work as expected.");
}

UCS2 *Parser::UCS2_strdup(const UCS2 *s)
{
    UCS2 *New;

    New=reinterpret_cast<UCS2 *>(POV_MALLOC((UCS2_strlen(s)+1) * sizeof(UCS2), UCS2toSysString(s).c_str()));
    UCS2_strcpy(New,s);
    return (New);
}

}
// end of namespace pov_parser
